--
-- Scratchpad MMU - Core VHDL module
-- Copyright (C) Jack Whitham 2009
-- http://www.jwhitham.org.uk/c/smmu.html
--
-- This library is free software; you can redistribute it and/or
-- modify it under the terms of the GNU Lesser General Public
-- License as published by the Free Software Foundation
-- (version 2.1 of the License only).
-- 
-- This library is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- Lesser General Public License for more details.
-- 
-- You should have received a copy of the GNU Lesser General Public
-- License along with this library; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
-- 

library ieee;
library unisim ;

use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned."+";
use ieee.std_logic_unsigned."-";
use ieee.std_logic_unsigned.conv_integer;
use unisim.vcomponents.all ;


entity smmu_core is
    -- Configuration
    generic (
        CPU_LEFT       : Natural ;
        SP_LEFT        : Natural ;
        DATA_LEFT      : Natural ;
        REF_LEFT       : Natural ;
        ACTIVE_LEFT    : Natural ) ;
    port (
        -- CPU side - really these are external ports
        cpu_address         : in std_logic_vector ( CPU_LEFT downto 0 );
        cpu_data_in         : in std_logic_vector ( DATA_LEFT downto 0 );
            -- cpu_data_in is also used for "size" and "reference in"
        cpu_data_out        : out std_logic_vector ( DATA_LEFT downto 0 );
            -- cpu_data_out is also used for "reference out"
        cpu_copy_loc        : in std_logic_vector ( SP_LEFT downto 0 );
        cpu_size            : in std_logic_vector ( SP_LEFT downto 0 );
        cpu_ben             : in std_logic_vector ( 3 downto 0 );
        cpu_write           : in std_logic; -- in 
        cpu_read            : in std_logic; -- in 
        cpu_open            : in std_logic; -- in 
        cpu_close           : in std_logic; -- in 
        cpu_ack             : out std_logic; -- out 
        -- Add Data LMB interface here
        -- Control lines from smmu to IPIF
        mst_cntl_rd_req     : out std_logic ;
        mst_cntl_wr_req     : out std_logic ;
        mst_cntl_bus_lock   : out std_logic ;
        mst_cntl_burst      : out std_logic ;
        mst_ip2bus_addr     : out std_logic_vector ( CPU_LEFT downto 0 ) ;
        mst_ip2bus_be       : out std_logic_vector ( 15 downto 0 ) ;
        mst_xfer_length     : out std_logic_vector ( 11 downto 0 ) ;
        mst_go              : out std_logic ;
        -- Status lines from IPIF to smmu
        mst_cmd_sm_set_done : in std_logic ;
        mst_cmd_sm_set_error : in std_logic ;
        mst_cmd_sm_set_timeout : in std_logic ;
        mst_cmd_sm_busy     : in std_logic ;
        mst_cmd_sm_clr_go   : in std_logic ;
        -- Data interface
        output_next         : in std_logic ;
        input_next          : in std_logic ;
        data_output         : out std_logic_vector ( DATA_LEFT downto 0 ) ;
        data_input          : in std_logic_vector ( DATA_LEFT downto 0 ) ;
        -- Only service addresses matching this:
        match_mask          : in std_logic_vector ( CPU_LEFT 
                                downto ACTIVE_LEFT + 1 ) ;
        -- Clock
        clk                 : in std_logic ;
        reset               : in std_logic ) ;
end entity smmu_core ;


architecture basic of smmu_core is


    -- Bus types
    subtype Bus_CPU_Address is std_logic_vector ( CPU_LEFT downto 0 ) ;
    subtype Bus_CPU_Data is std_logic_vector ( DATA_LEFT downto 0 ) ;
    subtype Bus_SP_Address is std_logic_vector ( SP_LEFT downto 0 ) ;

    -- Internal control lines from state machine
    signal open_operation   : std_logic;
    signal close_operation  : std_logic; 
    signal trigger          : std_logic; 
    signal next_ack         : std_logic;
    signal bus_side_write   : std_logic; 
    signal mst_go_int       : std_logic;
    signal bus_side_cen_override: std_logic;

    -- Other control lines
    signal sp_cpu_valid     : std_logic; -- delayed hit
    signal cpu_side_write   : std_logic_vector ( 3 downto 0 ) ;
    signal match_delay      : std_logic; -- delayed match
    signal match_inhibit    : std_logic;
    signal bus_side_cen     : std_logic;

    -- Status lines
    signal hit              : std_logic;
    signal miss             : std_logic;
    signal match            : std_logic;
    signal busy             : std_logic;
    signal some_words_remain: std_logic;
    signal one_word_remains : std_logic;

    -- Internal buses
    signal cpu_side_data_out: Bus_CPU_Data ;
    signal slow_data_out    : Bus_CPU_Data ;
    signal cpu_side_data_in : Bus_CPU_Data ;
    signal bus_side_data_out: Bus_CPU_Data ;
    signal bus_side_data_in : Bus_CPU_Data ;
    signal match_in_address : Bus_CPU_Address ;
    signal match_out_address: Bus_CPU_Address ;

    -- Internal registers
    signal data_input_delay : Bus_CPU_Data ;
    signal bus_side_address : Bus_SP_Address ;
    signal words_remaining  : Bus_SP_Address ;
    signal match_address    : Bus_CPU_Address ;
    signal saved_data_in    : Bus_CPU_Data ;
    signal slow_ack         : std_logic; 
    signal saved_be         : std_logic_vector ( 3 downto 0 ) ;

    -- The table
    constant NUM_ENTRIES    : Natural := 2 ** ( REF_LEFT + 1 ) ;

    signal ref              : Natural range 0 to NUM_ENTRIES - 1 ;

    type TABLE_ENTRY is record
                in_use      : std_logic ;
                min_address : Bus_CPU_Address ;
                max_address : Bus_CPU_Address ;
                offset      : Bus_CPU_Address ;
                out_address : Bus_CPU_Address ;
            end record ;

    type TABLE_ARRAY is array (0 to NUM_ENTRIES-1) of TABLE_ENTRY;

    signal table            : TABLE_ARRAY ;

    -- The state machine
    type State_Type is ( 
            READY , SLOW_ACCESS_WAIT , 
            BEGIN_OPEN , STAGE1_OPEN , STAGE2_OPEN , STAGE3_OPEN ,
            BEGIN_CLOSE , STAGE1_CLOSE , STAGE2_CLOSE , STAGE3_CLOSE ,
            HANDLE_TABLE_FULL , HANDLE_ERROR , HANDLE_TIMEOUT ,
            HANDLE_NOT_FOUND ) ;
    
    signal state            : State_Type ;

begin 

    -- Comparator network identifies "hits" on the SMMU
    process ( match_in_address , table , ref , 
                    close_operation , match_inhibit ) is
    begin
        match <= '0' ;
        match_out_address <= table(0).out_address;
        for i in 0 to ( NUM_ENTRIES - 1 )
        loop
            -- Comparison and out_address computation
            -- are done in parallel
            if (( table(i).in_use = '1' )
            and ( table(i).min_address <= match_in_address )
            and ( table(i).max_address > match_in_address )
            and (( close_operation = '0' ) or ( i < ref )))
            then
                match <= not match_inhibit ;
                match_out_address <= table(i).out_address;
            end if;
        end loop;
    end process;

    process ( match_in_address , table ) is
    begin
        for i in 0 to ( NUM_ENTRIES - 1 )
        loop
            -- out_address computation
            table(i).out_address <= match_in_address + table(i).offset;
            table(i).out_address ( 1 downto 0 ) <= "00" ;
            table(i).out_address ( CPU_LEFT downto SP_LEFT + 3 )
                                    <= ( others => '0' ) ;
        end loop ;
    end process ;

    process ( open_operation , close_operation , match_mask ,
                cpu_address , match_address ) is
    begin
        -- match_in_address multiplexer
        match_inhibit <= '0' ;
        if (( open_operation or close_operation ) = '1' )
        then
            -- OPEN/CLOSE operation
            match_in_address <= match_address ;
        else
            -- Regular access from CPU
            match_in_address <= cpu_address ;
            if ( cpu_address ( 
                    CPU_LEFT downto ACTIVE_LEFT + 1 ) /= match_mask )
            then
                match_inhibit <= '1' ;
            end if ;
        end if ;
        -- Mask out unused bits to reduce comparator size
        match_in_address ( 1 downto 0 ) <= "00" ;
        match_in_address ( CPU_LEFT downto ACTIVE_LEFT + 1 ) 
                                <= ( others => '0' ) ;
    end process;

    -- "Hits" are acknowledged on the following clock cycle.
    -- "Misses" are serviced by the memory access mechanism.
    hit <= match and ( cpu_read or cpu_write ) ;
    miss <= ( not match ) and ( cpu_read or cpu_write ) ;
    cpu_ack <= sp_cpu_valid or slow_ack;
    cpu_data_out <= cpu_side_data_out when ( sp_cpu_valid = '1' )
                else slow_data_out;
    cpu_side_data_in <= cpu_data_in when (( hit and cpu_write ) = '1' )
                else bus_side_data_out ;
    bus_side_data_in <= 
            data_input_delay 
                when (( open_operation and not match_delay ) = '1' )
            else cpu_side_data_out ;
    busy <= mst_cmd_sm_busy or mst_go_int or trigger ;
    some_words_remain <= '1' when ( conv_integer ( words_remaining ) /= 0 )
                else '0' ;
    one_word_remains <= '1' when ( conv_integer ( words_remaining ) = 1 )
                else '0' ;
    data_output <= bus_side_data_out when ( close_operation = '1' )
                else saved_data_in ;
    bus_side_cen <= bus_side_cen_override
            or ( close_operation and some_words_remain and output_next ) ;

    -- Outgoing controls to PLB IPIF
    mst_cntl_bus_lock <= '0' ;
    mst_go <= mst_go_int ;
    mst_xfer_length <= 
        -- 1 word if a "slow" operation:
        x"004" when (( open_operation or close_operation ) = '0' )
        -- -- 1 word if words_remaining = 0
        -- else x"004" when ( some_words_remain = '0' ) -- unnecessary?
        -- 16 words if words_remaining > 16
        else x"040" when ( conv_integer ( 
                        words_remaining ( SP_LEFT downto 4 ) ) /= 0 )
        -- otherwise, words_remaining 
        else "000000" & words_remaining ( 3 downto 0 ) & "00" ;
    mst_cntl_burst <= ( open_operation or close_operation ) 
            and some_words_remain and not one_word_remains ;


    lane_gen : for lane in 0 to 3
    generate 
        lane_block : block
            constant right  : Natural := 8 * lane ;
            constant left   : Natural := right + 7 ;
        begin
            -- Control for this byte lane
            cpu_side_write ( lane ) <= 
                        ( hit and cpu_write and cpu_ben ( lane )) 
                    or ( close_operation and output_next 
                        and some_words_remain and match ) ;

            copy : for i in 0 to 3
            generate
                mst_ip2bus_be ( lane + ( i * 4 )) <= saved_be ( lane ) ;
            end generate copy ;

            -- Memory for this byte lane
            scratchpad : RAMB16_S9_S9
                port map (
                    addrA => match_out_address ( SP_LEFT + 2 downto 2 ),
                    diA => cpu_side_data_in ( left downto right ),
                    dipA => "0",
                    doA => cpu_side_data_out ( left downto right ),
                    dopA => open,
                    clkA => clk,
                    enA => '1',
                    ssrA => reset,
                    weA => cpu_side_write ( lane ),

                    addrB => bus_side_address,
                    diB => bus_side_data_in ( left downto right ),
                    dipB => "0",
                    doB => bus_side_data_out ( left downto right ),
                    dopB => open,
                    clkB => clk,
                    enB => bus_side_cen,
                    ssrB => reset,
                    weB => bus_side_write);

        end block lane_block ;
    end generate lane_gen ;

    -- State machine
    process ( clk , reset ) is
    begin
        if ( reset = '1' )
        then
            state <= READY;
            for i in 0 to NUM_ENTRIES - 1
            loop
                table(i).in_use <= '0' ;
                table(i).min_address <= ( others => '0' ) ;
                table(i).max_address <= ( others => '0' ) ;
                table(i).offset <= ( others => '0' ) ;
            end loop;
            mst_go_int <= '0' ;
            sp_cpu_valid <= '0' ;
            slow_data_out <= x"e0e0e000" ;

        elsif ( clk'event )
        and ( clk = '1' )
        then

            slow_ack <= next_ack ;
            next_ack <= '0' ;
            trigger <= '0' ;
            match_delay <= match ;
            sp_cpu_valid <= hit;
            bus_side_write <= '0' ;
            bus_side_cen_override <= '0' ;

            case state is
            when READY =>
                    mst_cntl_rd_req <= '0' ;
                    mst_cntl_wr_req <= '0' ;
                    open_operation <= '0' ;
                    close_operation <= '0' ;
                    match_address <= cpu_address ;
                    saved_data_in <= cpu_data_in ;
                    words_remaining <= cpu_size ;
                    mst_ip2bus_addr <= cpu_address ;
                    saved_be <= cpu_ben ;
                    ref <= conv_integer ( 
                            cpu_data_in ( REF_LEFT downto 0 ) ) ;

                    if ( cpu_open = '1' )
                    then
                        state <= BEGIN_OPEN ;
                    elsif ( cpu_close = '1' )
                    then
                        state <= BEGIN_CLOSE ;
                    elsif ( miss = '1' )
                    then
                        -- Slow load or slow store; service
                        -- through main memory. Set up the 
                        -- transaction now.
                        mst_cntl_rd_req <= not cpu_write ;
                        mst_cntl_wr_req <= cpu_write ;
                        trigger <= '1' ;
                        state <= SLOW_ACCESS_WAIT ;
                    end if;

            when SLOW_ACCESS_WAIT =>
                    ref <= NUM_ENTRIES - 1;
                    if ( input_next = '1' )
                    then
                        slow_data_out <= data_input ;
                    end if ;
                    if ( busy = '0' )
                    then
                        next_ack <= '1' ;
                        state <= READY ;
                    end if ;

            when BEGIN_OPEN =>
                    state <= HANDLE_TABLE_FULL;
                    for i in 0 to NUM_ENTRIES - 1
                    loop
                        if ( table(i).in_use = '0' )
                        then
                            ref <= i;
                            state <= STAGE1_OPEN ;
                        end if;
                    end loop;
                    open_operation <= '1' ;

            when STAGE1_OPEN =>
                    table(ref).min_address <= match_address ;
                    table(ref).max_address <= match_address + 
                                ( words_remaining & "00" ) ;
                    table(ref).offset <= ( cpu_copy_loc & "00" ) - 
                                                    match_address ;
                    bus_side_address <= cpu_copy_loc - 1 ;
                    slow_data_out <= ( others => '0' ) ;
                    slow_data_out ( REF_LEFT downto 0 ) <= 
                        conv_std_logic_vector ( ref , REF_LEFT + 1 ) ;
                    state <= STAGE2_OPEN ;

            when STAGE2_OPEN =>
                    mst_cntl_rd_req <= '1' ;
                    saved_be <= x"f" ;
                    mst_ip2bus_addr <= match_address ;
                    trigger <= '1' ;
                    bus_side_cen_override <= '1' ;
                    state <= STAGE3_OPEN ;

            when STAGE3_OPEN =>
                    if ( busy = '0' )
                    then
                        if ( some_words_remain = '1' )
                        then
                            state <= STAGE2_OPEN ;
                        else
                            table(ref).in_use <= '1' ;
                            next_ack <= '1' ;
                            state <= READY ;
                        end if ;
                    end if ;

                    if (( input_next = '1' )
                    and ( some_words_remain = '1' ))
                    then
                        bus_side_write <= '1' ;
                        data_input_delay <= data_input ;
                        match_address <= match_address + 4 ;
                        bus_side_address <= 
                            bus_side_address + 1 ;
                        words_remaining <= 
                            words_remaining - 1 ;
                    end if ;
                    bus_side_cen_override <= '1' ;
            
            when BEGIN_CLOSE =>
                    if ( table(ref).in_use = '0' )
                    then
                        state <= HANDLE_NOT_FOUND ;
                    else
                        match_address <= table(ref).min_address ;
                        words_remaining <= ( 
                            table(ref).max_address ( SP_LEFT + 2 downto 2 ) -
                            table(ref).min_address ( SP_LEFT + 2 downto 2 ) ) ;
                        bus_side_address <= ( 
                            table(ref).offset ( SP_LEFT + 2 downto 2 ) +
                            table(ref).min_address ( SP_LEFT + 2 downto 2 ) ) ;
                        state <= STAGE1_CLOSE ;
                    end if;
                    close_operation <= '1' ;
       
            when STAGE1_CLOSE =>
                    state <= STAGE2_CLOSE ;
                    -- ensure that first word is available on the
                    -- block RAM output
                    bus_side_cen_override <= '1' ;

            when STAGE2_CLOSE =>
                    bus_side_address <= bus_side_address + 1 ;
                    mst_cntl_wr_req <= '1' ;
                    saved_be <= x"f" ;
                    mst_ip2bus_addr <= match_address ;
                    trigger <= '1' ;
                    state <= STAGE3_CLOSE ;

            when STAGE3_CLOSE =>
                    if ( output_next = '1' )
                    then
                        if ( some_words_remain = '1' )
                        then
                            match_address <= match_address + 4 ;
                            bus_side_address <= 
                                bus_side_address + 1 ;
                            words_remaining <= 
                                words_remaining - 1 ;
                        end if ;
                    elsif ( busy = '0' )
                    then
                        if ( some_words_remain = '1' )
                        then
                            state <= STAGE2_CLOSE ;
                            bus_side_address <= bus_side_address - 1 ;
                        else
                            table(ref).in_use <= '0' ;
                            next_ack <= '1' ;
                            state <= READY ;
                        end if ;
                    end if ;

            when HANDLE_NOT_FOUND =>
                    slow_data_out <= x"e0e0e013" ;
                    next_ack <= '1' ;
                    state <= READY ;

            when HANDLE_TABLE_FULL =>
                    slow_data_out <= x"e0e0e010" ;
                    next_ack <= '1' ;
                    state <= READY ;

            when HANDLE_ERROR =>
                    if ( busy = '0' )
                    then
                        slow_data_out <= x"e0e0e011" ;
                        next_ack <= '1' ;
                        state <= READY ;
                    end if ;

            when HANDLE_TIMEOUT =>
                    if ( busy = '0' )
                    then
                        slow_data_out <= x"e0e0e012" ;
                        next_ack <= '1' ;
                        state <= READY ;
                    end if ;

            end case ;

            -- Additional control lines
            if ( trigger = '1' )
            then
                mst_go_int <= '1' ;
            end if ;
            if ( mst_cmd_sm_clr_go = '1' )
            then
                mst_go_int <= '0' ;
            end if ;

            if ( mst_cmd_sm_set_timeout = '1' )
            then
                state <= HANDLE_TIMEOUT ;
            elsif ( mst_cmd_sm_set_error = '1' )
            then
                state <= HANDLE_ERROR ;
            end if ;

            -- Replace some parts of the table with zero logic
            -- to make the comparators smaller
            for i in 0 to NUM_ENTRIES - 1
            loop
                table(i).min_address ( 1 downto 0 ) <= "00" ;
                table(i).max_address ( 1 downto 0 ) <= "00" ;
                table(i).offset ( 1 downto 0 ) <= "00" ;
                table(i).min_address ( CPU_LEFT downto ACTIVE_LEFT + 1 )
                        <= ( others => '0' ) ;
                table(i).max_address ( CPU_LEFT downto ACTIVE_LEFT + 1 )
                        <= ( others => '0' ) ;
                table(i).offset ( CPU_LEFT downto ACTIVE_LEFT + 1 )
                        <= ( others => '0' ) ;
            end loop;
            match_address ( 1 downto 0 ) <= "00" ;
            match_address ( CPU_LEFT downto ACTIVE_LEFT + 1 ) <= match_mask ;

        end if;
    end process;


end architecture basic ;

