 
# Virtual Lab - optional components - unbuffered input reader for Twisted
# $Id: terminput.py,v 1.1 2009/08/11 20:52:38 jack Exp jack $
#
# Copyright (C) 2009, Jack Whitham
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import sys, os, atexit

from twisted.internet import reactor, interfaces


def __Init_Module():
    """Configure the standard input for "raw mode", where the input
    is unbuffered and special characters (^C, ^Z) are taken literally.
    """
    
    global RAW_MODE, UNIX_RESTORE, TRANSPORT
    if ( RAW_MODE ):
        return

    UNIX_RESTORE = None
    try:
        import msvcrt

        # Windows mode
        reactor.callLater(0, __WinReaderFunction)
    except ImportError:
        try:
            import tty, termios, fcntl

            # Unix mode, do some setup
            UNIX_RESTORE = termios.tcgetattr(STDIN_FD)
            tty.setraw(STDIN_FD)
            fcntl.fcntl(STDIN_FD, fcntl.F_SETFL, os.O_NONBLOCK)

            # Start reader class
            reactor.addReader(__UnixReaderClass())

        except ImportError:
            raise Exception("Raw mode input isn't supported because "
                "the following modules can't be imported: "
                "msvcrt (Windows) or tty, termios, fcntl (Unix)")

    # __Line_Mode is called on exit
    RAW_MODE = True
    atexit.register(__Line_Mode)

    
def __Line_Mode():
    """Return the standard input to the usual mode if necessary.
    This function is called automatically on exit if RawMode was
    ever called."""

    global RAW_MODE, UNIX_RESTORE
    if ( RAW_MODE and ( UNIX_RESTORE != None )):
        import termios, fcntl

        # Undo the setup
        termios.tcsetattr(STDIN_FD, termios.TCSADRAIN, UNIX_RESTORE)
        fcntl.fcntl(STDIN_FD, fcntl.F_SETFL, 0)
        UNIX_RESTORE = None

class __UnixReaderClass:
    def __init__(self):
        pass
            
    def fileno(self):
        return STDIN_FD

    def doRead(self):
        c = sys.stdin.read(1)
        if (( len(c) == 1 )
        and ( TRANSPORT != None )):
            TRANSPORT.write(c)

    def logPrefix(self):
        return '__UnixReaderClass'

    def connectionLost(self, reason=None):
        pass

def __WinReaderFunction():
    # Poll for input 25 times per second.
    # Of course polling is not a good way to do this. Something like
    # __UnixReaderClass is the correct way, but this won't work on
    # Windows because you can't select on a file descriptor (like STDIN_FD).
    # Another way is to use a thread, but that doesn't fit in well with
    # Twisted and could also mean that the program will hang on exit,
    # waiting for getch() to return before the thread can be stopped.

    import msvcrt
    while ( msvcrt.kbhit() ):
        c = msvcrt.getch()
        if (( len(c) == 1 )
        and ( TRANSPORT != None )):
            TRANSPORT.write(c)

    reactor.callLater(0.04, __WinReaderFunction)

def setTransport(transport):
    """Register a Transport to receive bytes from the keyboard.

    The Transport's write() method will be called with a byte
    whenever the user presses a key."""

    global TRANSPORT, RAW_MODE

    if ( not RAW_MODE ):
        __Init_Module()

    if ( TRANSPORT != None ):
        try:
            TRANSPORT.loseConnection()
        except AttributeError:
            pass

    TRANSPORT = transport
    

STDIN_FD = sys.stdin.fileno()
RAW_MODE = False
UNIX_RESTORE = None
TRANSPORT = None


