/*
 * Scratchpad MMU - Base Pointers Analysis Module
 * Copyright (C) Jack Whitham 2009
 * http://www.jwhitham.org.uk/c/smmu.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * (version 2.1 of the License only).
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>
#include <ctype.h>
#include <arpa/inet.h>

#define MAX_CALL_STACK_SIZE     128
#define STACK_SIZE_LIMIT        ( 1 << 20 ) /* 1 megaword */
#define SAMPLE_LIST_SIZE        7
#define SCRATCHPAD_ADDRESS      0xc0000000UL

#define FL_SET_D                0x1
#define FL_STOP                 0x2
#define FL_SLOW                 0x4
#define FL_LOAD                 0x8
#define FL_STORE                0x10
#define FL_DUMP                 0x20
#define FL_OPEN                 0x40
#define FL_CLOSE                0x80

#include "mb_mem.h"
#include "mb_elf.h"
#include "smmu.h"
#include "base_ptrs.h"
#include "settings.h"

typedef struct Sample_Record_struct
{
    unsigned    size ;
    unsigned    sample [ SAMPLE_LIST_SIZE ] ;
} Sample_Record ;


typedef struct Base_Pointer_struct
{
    unsigned    context ;
    unsigned    pc ;
    unsigned    base ;
    int         min_offset , max_offset ;
    unsigned    load_count ;
    unsigned    store_count ;
    unsigned    create_count ;
    unsigned    open_load_count ;
    unsigned    open_store_count ;
    unsigned    slow_count ;
    unsigned    fast_count ;
    unsigned    miss_count ;
    int         dump_me , open_now , am_ever_open ;
    int         open_at_dump ;
    int         score ;
    Sample_Record   user_sample ;
    Sample_Record   miss_user_sample ;
    unsigned    open_ref ;
    char        comment [ COMMENT_SIZE ] ;
    struct Base_Pointer_struct * hash_chain ;
} Base_Pointer ;

typedef struct SMMU_Data_struct {
    Base_Pointer *  bp ;
    Base_Pointer *  bp_alias ;
    int             open_now ;
} SMMU_Data ;

typedef struct BP_Context_struct 
{
    unsigned    flags ;
    unsigned    num_entries ;
    unsigned    sp_size ;
    unsigned    latency ;
    unsigned    ea ;
    unsigned    call_context ;
    unsigned    call_stack [ MAX_CALL_STACK_SIZE ] ;
    unsigned    call_stack_size ;
    unsigned    dump_command ;
    unsigned    activate_smmu ;
    Base_Pointer ** stack_base_ptr ;
    unsigned    stack_base_ptr_size ;
    unsigned    stack_base_ptr_dirty ;
    Base_Pointer ** bp_list ;
    Base_Pointer ** table ;
    unsigned    table_size ;
    unsigned    bp_count ;
    unsigned    hit_ref ;
    int         collapse_contexts ;
    MB_Context * mc1 ;
    MB_Context * mc2 ;
    struct SMMU_Context_struct * sc1 ;
    struct SMMU_Context_struct * sc2 ;
} BP_Context ;

typedef enum { MODE_ORDER_BY_MISS ,
            MODE_ORDER_BY_HEURISTIC } BP_Search_Mode ;

static Base_Pointer * Hash_Alloc ( BP_Context * bc , 
                unsigned context , unsigned pc )
{
    unsigned        key = context % bc -> table_size ;
    unsigned        length = 0 ;
    Base_Pointer *  found ;

    for ( found = bc -> table [ key ] ; 
                found != NULL ;
                found = found -> hash_chain )
    {
        if (( found -> context == context )
        && ( found -> pc == pc ))
        {
            break ;
        }
        length ++ ;
    }
    if ( found == NULL )
    {
        found = calloc ( 1 , sizeof ( Base_Pointer ) ) ;
        found -> context = context ;
        found -> pc = pc ;
        found -> open_ref = UINT_MAX ;
        found -> hash_chain = bc -> table [ key ] ;
        bc -> table [ key ] = found ;
        bc -> bp_count ++ ;
    }
    if ( length > 10 )
    {
        /* Rehash */
        unsigned new_table_size = bc -> table_size * 2 ;
        unsigned i , key ;
        Base_Pointer ** new_table = calloc ( new_table_size , 
                            sizeof ( Base_Pointer * ) ) ;
        Base_Pointer * j ;
        Base_Pointer * hash_chain ;

        for ( i = 0 ; i < bc -> table_size ; i ++ )
        {
            for ( j = bc -> table [ i ] ; 
                        j != NULL ;
                        j = hash_chain )
            {
                key = j -> context % new_table_size ;
                hash_chain = j -> hash_chain ;
                j -> hash_chain = new_table [ key ] ;
                new_table [ key ] = j ;
            }
        }
        free ( bc -> table ) ;
        bc -> table = new_table ;
        bc -> table_size = new_table_size ;
    }
    return found ;
}

static unsigned Combine_Context ( unsigned context , unsigned pc )
{
    unsigned i ;
    unsigned bits = 32 ;
    unsigned poly = 0xEDB88320 ;
    unsigned top_bit = ( 1U << ( bits - 1U )) ;

    for ( i = 0 ; i < bits ; i ++ )
    {
        unsigned do_xor = ((( context ^ ( pc << i )) & top_bit ) != 0U ) ;

        context = context << 1U ;
        if ( do_xor )
        {
            context ^= poly ;
        }
    }
    return context ;
}

static void BP_Trace_1 ( void * t_user , MB_Context * mc , 
                            MB_Trace_Name trace_name , const void * param )
{
    BP_Context * bc = (BP_Context *) t_user ;
    int recompute_context = 0 ;
    unsigned param_u = (unsigned) ( (uintptr_t) param ) ;

    switch ( trace_name )
    {
        case MB_LOAD :
                bc -> ea = param_u ;
                bc -> flags |= FL_LOAD ;
                break ;
        case MB_STORE :
                bc -> ea = param_u ;
                bc -> flags |= FL_STORE ;
                break ;
        case MB_OPEN :
                bc -> flags |= FL_OPEN ;
                break ;
        case MB_CLOSE :
                bc -> flags |= FL_CLOSE ;
                break ;
        case MB_HIT_LOAD :
        case MB_HIT_STORE :
                bc -> hit_ref = param_u ;
                break ;
        case MB_MISS_LOAD :
        case MB_MISS_STORE :
                bc -> flags |= FL_SLOW ;
                break ;
        case MB_SET_D :
                bc -> flags |= FL_SET_D ;
                break ;
        case MB_SIM_CMD :
                if ( param_u == bc -> dump_command )
                {
                    bc -> flags |= FL_DUMP ;
                } else if ( param_u == 1 )
                {
                    bc -> flags |= FL_STOP ;
                }
                break ;
        case MB_CALL :
                /* printf ( "MB_CALL to pc 0x%x" , param_u ) ; */
                assert ( bc -> call_stack_size < MAX_CALL_STACK_SIZE ) ;
                bc -> call_stack [ bc -> call_stack_size ] = mc -> cur_pc ;
                bc -> call_stack_size ++ ;
                recompute_context = 1 ;
                break ;
        case MB_RETURN :
                /* printf ( "MB_RETURN to pc 0x%x" , param_u ) ; */
                assert ( bc -> call_stack_size > 0 ) ;
                bc -> call_stack_size -- ;
                recompute_context = 1 ;
                break ;
        default :
                break ;
    }
    if (( bc -> flags & ( FL_LOAD | FL_STORE | FL_SLOW ))
    && (( bc -> ea & 0xffff0000UL ) == SCRATCHPAD_ADDRESS ))
    {
        /* Static RAM access - low latency in simulator.
         * This is where the SMMU allocation data is kept.
         */
        bc -> mc1 -> clock_cycle -= bc -> latency ;
        bc -> mc2 -> clock_cycle -= bc -> latency ;
        bc -> flags &= ~ ( FL_LOAD | FL_STORE | FL_SLOW ) ;
    }
    if ( recompute_context )
    {
        if ( bc -> collapse_contexts )
        {
            bc -> call_context = mc -> cur_pc ;
        } else {
            unsigned        i ;

            bc -> call_context = 0 ;
            for ( i = 0 ; i < bc -> call_stack_size ; i ++ )
            {
                bc -> call_context = Combine_Context (
                            bc -> call_context , 
                            bc -> call_stack [ i ] ) ;
            }
        }
    }
}

static void BP_Trace_2 ( void * t_user , MB_Context * mc , 
                            MB_Trace_Name trace_name , const void * param )
{
    /*MB_Context * mc2 = (MB_Context *) t_user ;*/
}

#define XUN_LSR_TX_BUFFER_EMPTY     0x20 /* Transmit holding reg empty */
#define XUN_THR_OFFSET  0x03 /* Transmit holding register */

static unsigned UART_Load ( void * io_user , unsigned address , unsigned size )
{
    return XUN_LSR_TX_BUFFER_EMPTY ; 
}

static void UART_Store_1 ( void * io_user , unsigned address , 
                            unsigned data , unsigned size )
{
    address &= 0x1f ;
    if (( address == XUN_THR_OFFSET )
    && ( isprint ( data ) || ( data == '\n' )))
    {
        fputc ( (char) data , stdout ) ;
        fflush ( stdout ) ;
    }
}

static void UART_Store_2 ( void * io_user , unsigned address , 
                            unsigned data , unsigned size )
{
}

static int BP_Score_Compare ( const void * p1 , const void * p2 )
{
    const Base_Pointer * bp1 = (* ((const Base_Pointer **) p1));
    const Base_Pointer * bp2 = (* ((const Base_Pointer **) p2));

    if ( bp1 -> score < bp2 -> score )
    {
        return 1 ;
    } else if ( bp1 -> score > bp2 -> score )
    {
        return -1 ;
    } else if ( bp1 -> pc < bp2 -> pc )
    {
        return -1 ;
    } else if ( bp1 -> pc > bp2 -> pc )
    {
        return 1 ;
    } else {
        return 0 ;
    }
}

static void Sample ( Sample_Record * sr , unsigned pc )
{
    if ( sr -> size < SAMPLE_LIST_SIZE )
    {
        int found = 0 ;
        unsigned i ;

        for ( i = 0 ; i < sr -> size ; i ++ )
        {
            if ( sr -> sample [ i ] == pc )
            {
                found = 1 ;
                break ;
            }
        }
        if ( ! found )
        {
            sr -> sample [ sr -> size ] = pc ;
            sr -> size ++ ;
        }
    }
}

static void Print_Sample ( Sample_Record * sr )
{
    unsigned j ;

    for ( j = 0 ; j < sr -> size ; j ++ )
    {
        printf ( " %08x" , sr -> sample [ j ] ) ;
    }
    if ( sr -> size == SAMPLE_LIST_SIZE )
    {
        printf ( "..." ) ;
    }
}

static int Heur_Cost ( BP_Context * bc , Base_Pointer * bp )
{
    return (int) ( 2 * bp -> create_count * 
                      ( bc -> latency + (( bp -> max_offset 
                                        - bp -> min_offset ) / 4 ))) ;
}
static int Heur_Benefit ( BP_Context * bc , Base_Pointer * bp )
{
    return (int) (( bc -> latency - 1 ) * 
                        ( bp -> open_load_count + bp -> open_store_count )) ;
}

static void BP_Report ( 
            BP_Context * bc ,
            BP_Search_Mode mode )
{
    unsigned        bp_index , i ;

    /* Normal report header */
    switch ( mode )
    {
        case MODE_ORDER_BY_MISS :
                printf ( "* During OPEN section, base pointers in existence,\n" 
                         "  (ordered by miss count):\n" ) ;
                break ;
        case MODE_ORDER_BY_HEURISTIC :
                printf ( "* At dump point, base pointers in existence,\n" 
                         "  (ordered by score heuristic):\n" ) ;
                break ;
        default :
                assert ( 0 ) ;
                break ;
    }

    /* Reassemble base pointers into a list */
    bp_index = 0 ;

    for ( i = 0 ; i < bc -> table_size ; i ++ )
    {
        Base_Pointer * j ;
        for ( j = bc -> table [ i ] ; 
                    j != NULL ;
                    j = j -> hash_chain )
        {
            assert ( bp_index < bc -> bp_count ) ;
            switch ( mode )
            {
                case MODE_ORDER_BY_MISS :
                        j -> score = j -> miss_count ;
                        break ;
                case MODE_ORDER_BY_HEURISTIC :
                        j -> score = Heur_Benefit ( bc , j ) - 
                                    Heur_Cost ( bc , j ) ;
                        break ;
            }
            bc -> bp_list [ bp_index ] = j ;
            bp_index ++ ;
        }
    }
    assert ( bp_index == bc -> bp_count ) ;

    /* Sort the list by score (descending) */
    qsort ( bc -> bp_list , bc -> bp_count , sizeof ( Base_Pointer * ) ,
                    BP_Score_Compare ) ;

    /* Print out interesting base pointers (dump_me = 1) */
    for ( i = 0 ; i < bc -> bp_count ; i ++ )
    {
        Base_Pointer * bp = bc -> bp_list [ i ] ;
        int enable = 0 ;

        switch ( mode )
        {
            case MODE_ORDER_BY_MISS :
                    enable = ( bp -> miss_count != 0 ) ;
                    break ;
            case MODE_ORDER_BY_HEURISTIC :
                    enable = ( bp -> dump_me ) ;
                    break ;
        }
        if ( enable
        && (( bp -> load_count + bp -> store_count ) != 0 ))
        {
            printf ( "%d - created %08x: cc %u lc %u sc %u f %1.1f%%\n"
                    "  size %d (base %08x) eo:%u oad:%u " ,
                    bp -> score ,
                    bp -> pc , 
                    bp -> create_count ,
                    bp -> open_load_count ,
                    bp -> open_store_count ,
                    (double) ( bp -> fast_count * 100.0 ) /
                    (double) ( bp -> miss_count + bp -> fast_count ) ,
                    bp -> max_offset - bp -> min_offset ,
                    bp -> base ,
                    bp -> am_ever_open ,
                    bp -> open_at_dump ) ;

            if ( strlen ( bp -> comment ) != 0 )
            {
                    printf ( "id: %s" , bp -> comment ) ;
            }
            printf ( "\n  used at: " ) ;
            Print_Sample ( & bp -> user_sample ) ;
            printf ( "\n  miss at: " ) ;
            Print_Sample ( & bp -> miss_user_sample ) ;
            printf ( "\n\n" ) ;
        }
    }
}

static void Latex_Integer ( FILE * fd , int x )
{
    const char * suffix_table = "GMk " ;
    unsigned     divisor = 1000 * 1000 * 1000 ;
    unsigned     y , i ;

    /* Print 3 significant figures.
     * Start with giga suffix, then try mega, then kilo,
     * then single units. */
    fprintf ( fd , "&\t" ) ;
    if ( x < 0 )
    {
        x = - x ;
        fprintf ( fd , "-" ) ;
    }
    for ( i = 0 ; i < 4 ; i ++ )
    {
        y = x / divisor ;
        if (( y >= 100 )
        || ( i == 3 ))
        {
            /* 100+ or single units */
            fprintf ( fd , "%u%c" , y , suffix_table [ i ] ) ;
            return ;
        } else if ( y >= 10 )
        {
            /* 10+ - show 1 dp */
            fprintf ( fd , "%1.1f%c" , 
                    (double) x / (double) divisor , suffix_table [ i ] ) ;
            return ;
        } else if (( y >= 1 )
        && ( divisor > 1 ))
        {
            /* 1+ - show 2 dp */
            fprintf ( fd , "%1.2f%c" , 
                    (double) x / (double) divisor , suffix_table [ i ] ) ;
            return ;
        }
        /* try next suffix */
        divisor /= 1000 ;
    }
    assert ( 0 ) ;
}

static void BP_Latex_Report ( 
            BP_Context * bc ,
            const char * fname )
{
    FILE * fd ;
    unsigned i , j ;

    if ( bc -> bp_list == NULL ) return ;

#define EOTL "\\\\\n"
    fd = fopen ( fname , "wt" ) ;
    assert ( fd != NULL ) ;

    fprintf ( fd ,
           "\\begin{tabular}{|c|c|c|c|c|c|c|c|}\n"
           "\\hline\n"
"Label & \\multicolumn{3}{c|}{Counters}&Size &Cost     &Benefit    & Score " EOTL
"      & $l$ & $s$ & $c$               &$z$  & $cC(z)$ &$B(l+s)$   & " EOTL
           "\\hline\n" ) ;

    for ( i = 0 ; i < bc -> bp_count ; i ++ )
    {
        Base_Pointer * bp = bc -> bp_list [ i ] ;

        if (( ! bp -> dump_me )
        || (( bp -> open_load_count + bp -> open_store_count ) == 0 ))
        {
            continue ;
        }

        for ( j = 0 ; bp -> comment [ j ] != '\0' ; j ++ )
        {
            switch ( bp -> comment [ j ] )
            {
                case '_' :  fprintf ( fd , "\\_" ) ;
                            break ;
                case '>' :  fprintf ( fd , "$\\gt$" ) ;
                            break ;
                default :   fputc ( bp -> comment [ j ] , fd ) ;
                            break ;
            }
        }
        for ( ; j < 30 ; j ++ )
        {
            fputc ( ' ' , fd ) ;
        }
        fputc ( '\t' , fd ) ;
        Latex_Integer ( fd , bp -> open_load_count ) ;
        Latex_Integer ( fd , bp -> open_store_count ) ;
        Latex_Integer ( fd , bp -> create_count ) ;
        Latex_Integer ( fd , bp -> max_offset - bp -> min_offset ) ;
        Latex_Integer ( fd , Heur_Cost ( bc , bp ) ) ;
        Latex_Integer ( fd , Heur_Benefit ( bc , bp ) ) ;
        Latex_Integer ( fd , Heur_Benefit ( bc , bp ) - Heur_Cost ( bc , bp ) ) ;
        fprintf ( fd , EOTL ) ;
    }
    fprintf ( fd , "\\hline\n"
                "\\end{tabular}\n\n" ) ;
    fclose ( fd ) ;
}

static int Detect_Base_Pointers_Core ( 
            BP_Context * bc ,
            struct MB_System_Context_struct * sys1 ,
            struct MB_System_Context_struct * sys2 ,
            const char * fname1 ,
            const char * fname2 )
{
    const char *    out ;
    Base_Pointer *  reg_base_ptr [ 0x20 ] ;
    unsigned        i ;
    double          total ;
    MB_Context *    mc1 ;
    MB_Context *    mc2 ;
    SMMU_Data *     smmu_table_data ;
    Base_Pointer *  stack_bp = NULL ;
    unsigned        stack_base = 0 ;
    unsigned        open_count = 0 ;
    unsigned        total_stack_access_count = 0 ;
    unsigned        total_gh_access_count = 0 ;
    unsigned        total_miss_in_open = 0 ;
    unsigned        total_fast_in_open = 0 ;
    unsigned        total_stack_above_miss_in_open = 0 ;
    unsigned        open_begin = 0 ;
    unsigned        total_open = 0 ;
    int             warning = 0 ;
    unsigned        activate_smmu ;

    memset ( reg_base_ptr , 0 , 0x20 * sizeof ( Base_Pointer * ) ) ;

    out = MB_Read_Elf ( sys1 , fname1 ) ;
    if ( out != NULL )
    {
        printf ( "error reading %s: %s\n" , fname1 , out ) ;
        return 0 ;
    }
    out = MB_Read_Elf ( sys2 , fname2 ) ;
    if ( out != NULL )
    {
        printf ( "error reading %s: %s\n" , fname2 , out ) ;
        return 0 ;
    }

    /* Mechanism for passing parameters to simulated program: */
    activate_smmu = htonl (( bc -> activate_smmu & 0xff ) | 0x5e5e5e00 ) ;
    MB_Write ( sys1 , SCRATCHPAD_ADDRESS , 
                (const char *) & activate_smmu , 4 ) ;
    MB_Write ( sys2 , SCRATCHPAD_ADDRESS , 
                (const char *) & activate_smmu , 4 ) ;

    /* UART device for output */
    MB_Map_IO ( sys1 , 0x83e00000 >> PAGE_SHIFT , 
                    UART_Load , UART_Store_1 , bc ) ;
    MB_Map_IO ( sys2 , 0x83e00000 >> PAGE_SHIFT , 
                    UART_Load , UART_Store_2 , bc ) ;

    /* Contexts, SMMU initialisation */
    bc -> mc1 = mc1 = MB_System_Get_MB_Context ( sys1 ) ;
    bc -> mc2 = mc2 = MB_System_Get_MB_Context ( sys2 ) ;
    bc -> sc1 = SMMU_Init ( sys1 , bc -> num_entries , 0 , bc -> sp_size ) ;
    bc -> sc2 = SMMU_Init ( sys2 , bc -> num_entries , 0 , bc -> sp_size ) ;
    bc -> flags = 0 ;
    smmu_table_data = alloca ( bc -> num_entries * sizeof ( SMMU_Data ) ) ;
    memset ( smmu_table_data , 0 , 
                bc -> num_entries * sizeof ( SMMU_Data ) ) ;

    printf ( "Running program:\n" ) ;
    while ( ! ( bc -> flags & FL_STOP ))
    {
        unsigned        rD , rA , rB , d1 , d2 , context ;
        unsigned        rbp ;
        int             produces_pointer ;
        Base_Pointer *  bp ;
        Base_Pointer *  loaded_bp ;
        Base_Pointer *  address_bp = NULL ;

        if (( mc1 -> cur_pc != mc2 -> cur_pc )
        || (( mc1 -> cur_iword >> 16 ) != ( mc2 -> cur_iword >> 16 )))
        {
            printf ( "ERROR: desync: context 1 at pc %x, context 2 at pc %x\n"
                    "context 1 iword %x, context 2 iword %x\n" ,
                mc1 -> cur_pc , mc2 -> cur_pc ,
                mc1 -> cur_iword , mc2 -> cur_iword ) ;
            return 0 ;
        }
        bc -> flags = 0 ;
        bc -> hit_ref = UINT_MAX ;

        MB_Step ( mc1 , 0 ) ;
        MB_Step ( mc2 , 0 ) ;

        rA = Get_rA ( mc1 ) ;
        rB = Get_rB ( mc1 ) ;

        rD = Get_rD ( mc1 ) ;
        if ( bc -> flags & FL_SET_D )
        {
            d1 = Get_D ( mc1 ) ;
            d2 = Get_D ( mc2 ) ;
        } else {
            d1 = d2 = 0 ;
        }
        produces_pointer = ( d1 != d2 ) ;
        context = bc -> call_context ;
        bp = NULL ;
        rbp = 0 ;
        loaded_bp = NULL ;

        /* What base pointer acts as an input to this instruction? */
        if (( reg_base_ptr [ rA ] == NULL )
        && ( reg_base_ptr [ rB ] != NULL ))
        {
            /* Base pointer in rB. */
            bp = reg_base_ptr [ rB ] ;
            rbp = rB ;

        } else if (( reg_base_ptr [ rA ] != NULL )
        && ( reg_base_ptr [ rB ] == NULL ))
        {
            /* Base pointer in rA. */
            bp = reg_base_ptr [ rA ] ;
            rbp = rA ;
        }
        address_bp = bp ;
        /* How is the pointer, if any, used here? */
        if (( bc -> flags & ( FL_STORE | FL_LOAD ))
        && ( bp != NULL ))
        {
            /* Load/store to dynamic location */
            int offset = (int) ( bc -> ea - bp -> base ) ;
            if (( offset + 4 ) > bp -> max_offset )
            {
                bp -> max_offset = offset + 4 ;
            } 
            if ( offset < bp -> min_offset )
            {
                bp -> min_offset = offset ;
            } 
            if ( bc -> flags & FL_LOAD )
            {
                bp -> load_count ++ ;
            } else {
                bp -> store_count ++ ;
            }
            if ( open_count != 0 )
            {
                if ( bc -> flags & FL_LOAD )
                {
                    bp -> open_load_count ++ ;
                } else {
                    bp -> open_store_count ++ ;
                }
            }
            if ( stack_bp == bp )
            {
                total_stack_access_count ++ ;
            } else {
                total_gh_access_count ++ ;
            }
            if ( bc -> flags & FL_SLOW )
            {
                bp -> slow_count ++ ;
                if ( open_count != 0 )
                {
                    bp -> miss_count ++ ;
                    total_miss_in_open ++ ;
                    Sample ( & bp -> miss_user_sample , mc1 -> cur_pc ) ;
                }

                if ( bp -> open_now )
                {
                    /* We expected this base pointer to be "open" -
                     * but then we got this slow memory access.
                     */
                    unsigned    min , max , base ;
                    int         rc ;

                    assert ( open_count != 0 ) ;
                    rc = SMMU_Inspect_Table ( bc -> sc1 , 
                            bp -> open_ref , & min , & max , & base ,
                            NULL , NULL ) ;

                    if ( bp == stack_bp )
                    {
                        assert ( rc ) ;
                        assert ( min < max ) ;
                        if ( bc -> ea < min )
                        {
                            printf ( "\nERROR: part of the stack is open, "
            "but this access is below the open section.\n"
            "Open section is 0x%x:0x%x\n"
            "Access at pc 0x%x uses ea 0x%x\n" ,
                                min , max ,
                                mc1 -> cur_pc , bc -> ea ) ;
                            return 0 ;
                        } else {
                            total_stack_above_miss_in_open ++ ;
                        }
                    } else {
                        printf ( "\nERROR: base pointer r%u "
            "created at pc 0x%x with value 0x%x, is open -\n"
            "but access at pc 0x%x to ea 0x%x (offset %d) is a 'miss'.\n"
            "Comment: %s\n" ,
                            rbp , bp -> pc , bp -> base ,
                            mc1 -> cur_pc , bc -> ea , offset ,
                            bp -> comment ) ;
                        if ( rc )
                        {
                            printf ( "Base: 0x%x   Min: 0x%x  Max: 0x%x\n" ,
                                base , min , max ) ;
                        } else {
                            printf ( "Hmm? It's closed.\n" ) ;
                        }
                        SMMU_Dump_Table ( bc -> sc1 ) ;
                        return 0 ;
                    }
                }
                bp -> open_now = 0 ;
            } else if ( ! ( bc -> flags & ( FL_OPEN | FL_CLOSE )) )
            {
                unsigned    j ;
                int         rc ;

                assert ( open_count != 0 ) ;
                total_fast_in_open ++ ;
                bp -> fast_count ++ ;
                /* This access was serviced quickly =>
                 * the base pointer must be open.
                 * Actually this doesn't follow.. it might simply
                 * happen to overlap an open region.
                 */
                i = bc -> hit_ref ;
                j = bp -> open_ref ;
                assert ( i < bc -> num_entries ) ;

                if ( ! bp -> open_now )
                {
                    /* If it's not marked as "open" then it must be
                     * an alias of the real pointer */
                    j = bp -> open_ref = i ;
                    if ( smmu_table_data [ i ] . bp_alias != NULL )
                    {
                        smmu_table_data [ i ] . bp_alias -> open_now = 0 ;
                    }
                    smmu_table_data [ i ] . bp_alias = bp ;
                }

                assert ( j < bc -> num_entries ) ;
                assert ( smmu_table_data [ i ] . open_now ) ;
                assert ( smmu_table_data [ j ] . open_now ) ;

                /* Capture information about the SMMU region */
                if (( bp -> open_ref != bc -> hit_ref )
                && ( ! warning ))
                {
                    printf ( "\nWARNING: although the SMMU can "
                            "cope with pointer aliasing,\n"
        "the simulator has a simple mind and is not able to handle it.\n"
        "One base pointer has accessed both of the following SMMU entries:\n"
        "Entry %u: opened pc 0x%x with value 0x%x comment %s\n"
        "Entry %u: opened pc 0x%x with value 0x%x comment %s\n" 
        "Location: pc 0x%x\n" ,
                        i , smmu_table_data [ i ] . bp -> pc ,
                        smmu_table_data [ i ] . bp -> base ,
                        smmu_table_data [ i ] . bp -> comment ,
                        j , smmu_table_data [ j ] . bp -> pc ,
                        smmu_table_data [ j ] . bp -> base ,
                        smmu_table_data [ j ] . bp -> comment ,
                        mc1 -> cur_pc ) ;
                    SMMU_Dump_Table ( bc -> sc1 ) ;
                    warning = 1 ;
                }
                bp -> open_ref = bc -> hit_ref ;
                bp -> open_now = 1 ;
                bp -> am_ever_open = 1 ;
                rc = SMMU_Inspect_Table ( bc -> sc1 , 
                        bp -> open_ref , NULL , NULL , NULL , NULL ,
                        bp -> comment ) ;
                assert ( rc ) ;  /* must be open */
            }
            Sample ( & bp -> user_sample , mc1 -> cur_pc ) ;
        }
        /* Side effects of this operation - update the stack if necessary */
        if (( bc -> flags & ( FL_STORE | FL_LOAD ))
        && ( bc -> ea >= mc1 -> gpr [ 1 ] )
        && ( bc -> ea <= stack_base ))
        {
            /* Bona fide stack access */
            unsigned pos = ( stack_base / 4 ) - ( bc -> ea / 4 ) ;

            assert ( pos < STACK_SIZE_LIMIT ) ;
            if ( pos >= bc -> stack_base_ptr_size )
            {
                /* Increase the size of the array */
                Base_Pointer ** new_stack_base_ptr = calloc ( 
                        bc -> stack_base_ptr_size + pos , 
                        sizeof ( Base_Pointer * ) ) ;
                memcpy ( new_stack_base_ptr ,
                        bc -> stack_base_ptr ,
                        bc -> stack_base_ptr_size *
                            sizeof ( Base_Pointer * ) ) ;

                free ( bc -> stack_base_ptr ) ;

                bc -> stack_base_ptr_size += pos ;
                bc -> stack_base_ptr = new_stack_base_ptr ;
            }

            if ( bc -> flags & FL_STORE )
            {
                /* Write base pointer (or NULL) */
                bc -> stack_base_ptr [ pos ] = reg_base_ptr [ rD ] ;
                if (( pos + 1 ) > bc -> stack_base_ptr_dirty )
                {
                    bc -> stack_base_ptr_dirty = pos + 1 ;
                }
            } else {
                /* Read base pointer (or NULL) */
                loaded_bp = bc -> stack_base_ptr [ pos ] ;
                assert ( bc -> flags & FL_SET_D ) ;
            }
        } 
        /* The main effect of the operation - 
         * is a pointer produced as output? */
        if ( produces_pointer )
        {
            if ( bc -> flags & FL_LOAD )
            {
                /* New pointer from load */
                bp = loaded_bp ;
                if (( bp == NULL )
                && (( rA == 1 ) || ( rB == 1 )))
                {
                    assert ( 0 ) ;
                }
            } 
            
            if ( bp == NULL ) 
            {
                /* New pointer from load
                 * OR new pointer from "thin air"
                 * e.g. immediate */

                bp = Hash_Alloc ( bc , context ^ mc1 -> cur_pc ,
                            mc1 -> cur_pc ) ;
                bp -> create_count ++ ;
                bp -> base = d1 ;
            } else {
                /* Existing pointer is modified in
                 * some way e.g. offset changes */
            }
        } else {
            bp = NULL ;
        } 
        /* Update output register if necessary */
        if ( bc -> flags & FL_SET_D )
        {
            assert ( rD != 0 ) ;
            reg_base_ptr [ rD ] = bp ;
            if ( rD == 1 )
            {
                unsigned pos ;

                /* This is the stack pointer 
                 * (what else is r1 used for in C?) */
                if ( stack_base == 0 )
                {
                    stack_base = mc1 -> gpr [ 1 ] ;
                    stack_bp = bp ;
                    assert ( bp != NULL ) ;
                } else {
                    assert (( bp == NULL ) || ( bp == stack_bp )) ;
                }

                /* When r1 changes (due to removal of stack frames)
                 * we can expire some parts of the stack_base_ptr array. 
                 */
                pos = ( stack_base / 4 ) - ( mc1 -> gpr [ 1 ] / 4 ) ;

                assert ( pos < STACK_SIZE_LIMIT ) ;
                if ( bc -> stack_base_ptr_dirty > pos )
                {
                    for ( i = pos + 1 ; i < bc -> stack_base_ptr_dirty ; i ++ )
                    {
                        bc -> stack_base_ptr [ i ] = NULL ;
                    }
                    bc -> stack_base_ptr_dirty = pos + 1 ;
                }
            }
        }

        /* Further action for each optype */
        if ( bc -> flags & FL_DUMP )
        {
            /* Set dump_me flag on each base pointer
             * since each base pointer is active at the
             * point of interest. */
            for ( i = 1 ; i < 0x20 ; i ++ )
            {
                Base_Pointer * bp = reg_base_ptr [ i ] ;
                
                if ( bp == NULL ) continue ;

                if ( bp -> dump_me )
                {
                    bp -> open_at_dump = bp -> open_now && 
                        bp -> open_at_dump ;
                } else {
                    bp -> dump_me = 1 ;
                    bp -> open_at_dump = bp -> open_now ;
                }
            }
            for ( i = 0 ; i < bc -> stack_base_ptr_size ; i ++ )
            {
                Base_Pointer * bp = bc -> stack_base_ptr [ i ] ;
                
                if ( bp == NULL ) continue ;

                if ( bp -> dump_me )
                {
                    bp -> open_at_dump = bp -> open_now && 
                        bp -> open_at_dump ;
                } else {
                    bp -> dump_me = 1 ;
                    bp -> open_at_dump = bp -> open_now ;
                }
            }
        } else if ( bc -> flags & FL_OPEN )
        {
            int ok = 0 ;
            for ( i = 0 ; i < bc -> num_entries ; i ++ )
            {
                unsigned opened_at ;
                int exists = SMMU_Inspect_Table ( 
                        bc -> sc1 , 
                        i , NULL , NULL , NULL ,
                        & opened_at , NULL ) ;

                if (( exists )
                && ( ! smmu_table_data [ i ] . open_now )
                && ( opened_at == mc1 -> cur_pc ))
                {
                    /* Just opened! */
                    smmu_table_data [ i ] . bp = address_bp ;
                    smmu_table_data [ i ] . bp_alias = NULL ;
                    smmu_table_data [ i ] . open_now = 1 ;
                    assert ( address_bp != NULL ) ;
                    ok = 1 ;
                    break ;
                }
            }
            assert ( ok ) ;
            if ( open_count == 0 )
            {
                open_begin = mc1 -> clock_cycle ;
            }
            open_count ++ ;
        } else if ( bc -> flags & FL_CLOSE )
        {
            int ok = 0 ;
            for ( i = 0 ; i < bc -> num_entries ; i ++ )
            {
                int exists = SMMU_Inspect_Table ( 
                        bc -> sc1 , 
                        i , NULL , NULL , NULL ,
                        NULL , NULL ) ;

                if (( ! exists )
                && ( smmu_table_data [ i ] . open_now ))
                {
                    /* Just closed! */
                    if ( smmu_table_data [ i ] . bp != NULL ) 
                    {
                        smmu_table_data [ i ] . bp -> open_now = 0 ;
                        smmu_table_data [ i ] . bp -> open_ref = UINT_MAX ;
                    }
                    if ( smmu_table_data [ i ] . bp_alias != NULL )
                    {
                        smmu_table_data [ i ] . bp_alias -> open_now = 0 ;
                    }
                    smmu_table_data [ i ] . bp_alias = NULL ;
                    smmu_table_data [ i ] . bp = NULL ;
                    smmu_table_data [ i ] . open_now = 0 ;
                    ok = 1 ;
                    break ;
                }
            }
            assert ( ok ) ;/* assert that we found something */
            assert ( open_count != 0 ) ;
            open_count -- ;
            if ( open_count == 0 )
            {
                total_open += mc1 -> clock_cycle - open_begin ;
            }
        }
    }
    if ( open_count != 0 )
    {
        total_open += mc1 -> clock_cycle - open_begin ;
    }
    printf ( "\n\n" ) ;
    printf ( "\n\nReport:\n" ) ;
    printf ( "* Overall execution stats:\n"
                "  %u clock cycles.\n"
                "  %u accesses to stack.\n"
                "  %u accesses to heap and globals.\n\n" ,
            mc1 -> clock_cycle ,
            total_stack_access_count ,
            total_gh_access_count ) ;

    total = ( total_miss_in_open * bc -> latency ) + total_fast_in_open ;
    printf ( "* In section of interest (where something is open):\n"
                "  %u clock cycles (%1.1f%% of total cycles).\n"
                "  %u fixed execution cost (%1.1f%% of run cost).\n"
                "  %u accesses (%1.1f%% of run cost).\n"
                "  %u fast accesses (%1.1f%% of m.a. cost, %1.1f%% of count).\n"
                "  %u slow ('miss') accesses (%1.1f%% of m.a. cost, "
                        "%1.1f%% of count).\n"
                "  %u slow ('miss') accesses above the open stack region.\n\n" ,
            total_open ,
            ( 100.0 * (double) total_open ) / (double) mc1 -> clock_cycle ,
            (unsigned) ( total_open - total ) ,
            100.0 - (( 100.0 * (double) total ) / (double) total_open ) ,
            total_miss_in_open + total_fast_in_open ,
            ( 100.0 * (double) total ) / (double) total_open ,
            total_fast_in_open , 
                    100.0 * ( (double) total_fast_in_open / total ) ,
            100.0 * ( (double) total_fast_in_open / 
                        (double) ( total_miss_in_open + total_fast_in_open )),
            total_miss_in_open ,
                    100.0 * (( total - total_fast_in_open ) / total ) ,
            100.0 * ( 1.0 - ( (double) total_fast_in_open / 
                        (double) ( total_miss_in_open + total_fast_in_open ))),
            total_stack_above_miss_in_open ) ;

    bc -> bp_list = calloc ( bc -> bp_count , sizeof ( Base_Pointer * ) ) ;
    assert ( bc -> bp_list != NULL ) ;

    BP_Report ( bc , MODE_ORDER_BY_MISS ) ;
    BP_Report ( bc , MODE_ORDER_BY_HEURISTIC ) ;
    return 1 ;
}




static int Detect_Base_Pointers_For ( 
            const char * fname1 ,
            const char * fname2 ,
            unsigned activate_smmu ,
            unsigned dump_command ,
            const char * report_name ,
            int collapse_contexts ,
            int perfect )
{
    BP_Context *                        bc ;
    struct MB_System_Context_struct *   sys1 ;
    struct MB_System_Context_struct *   sys2 ;
    unsigned                            i ;
    int                                 rc ;

    printf ( "\n\n\n* Making report '%s'.\n"
            "  Activate SMMU setting 0x%x\n"
            "  Dump command is 0x%x\n\n" ,
                    report_name , activate_smmu ,
                    dump_command ) ;
            
    bc = calloc ( 1 , sizeof ( BP_Context ) ) ;
    assert ( bc != NULL ) ;
   
    /* Invariants that affect the simulated program */
    bc -> num_entries = SP_TABLE_SIZE ;
    bc -> sp_size = SP_SIZE ;
    bc -> latency = perfect ? 0 : EXT_MEM_LATENCY ;
    bc -> dump_command = dump_command ;
    bc -> activate_smmu = activate_smmu ;
    bc -> collapse_contexts = collapse_contexts ;

    /* Other settings that only affect simulator efficiency */
    bc -> table_size = 1024 ;
    bc -> stack_base_ptr_size = 1024;

    /* Allocate memory */
    bc -> table = calloc ( bc -> table_size , sizeof ( Base_Pointer * ) ) ;
    bc -> stack_base_ptr = calloc ( bc -> stack_base_ptr_size , 
                            sizeof ( Base_Pointer * ) ) ;
    assert ( bc -> table != NULL ) ;
    assert ( bc -> stack_base_ptr != NULL ) ;

    sys1 = MB_System_Init ( BP_Trace_1 , bc -> latency , bc ) ;
    sys2 = MB_System_Init ( BP_Trace_2 , bc -> latency , bc ) ;

    /* Go */
    rc = Detect_Base_Pointers_Core ( bc , 
                sys1 , sys2 , fname1 , fname2 ) ;
    BP_Latex_Report ( bc , report_name ) ;

    /* Free memory */
    if ( bc -> sc1 != NULL )
    {
        SMMU_Delete ( bc -> sc1 ) ;
    }
    if ( bc -> sc2 != NULL )
    {
        SMMU_Delete ( bc -> sc2 ) ;
    }
    MB_System_Delete ( sys2 ) ;
    MB_System_Delete ( sys1 ) ;

    for ( i = 0 ; i < bc -> table_size ; i ++ )
    {
        Base_Pointer * j ;
        Base_Pointer * k ;

        for ( j = bc -> table [ i ] ; 
                    j != NULL ;
                    j = k )
        {
            k = j -> hash_chain ;
            free ( j ) ;
        }
    }
    if ( bc -> bp_list != NULL )
    {
        free ( bc -> bp_list ) ;
    }
    free ( bc -> stack_base_ptr ) ;
    free ( bc -> table ) ;
    free ( bc ) ;
    return rc ;
}

void Detect_Base_Pointers ( 
            const char * fname1 ,
            const char * fname2 )
{
    if ( Detect_Base_Pointers_For ( fname1 , fname2 ,
            0 , 2 , "control.tex" , 0 , 0 ) 
    && Detect_Base_Pointers_For ( fname1 , fname2 ,
            8 | 3 , 4 , "dmcu.tex" , 0 , 0 ) 
    && Detect_Base_Pointers_For ( fname1 , fname2 ,
            0 , 2 , "perfect.tex" , 0 , 1 ) 
    && Detect_Base_Pointers_For ( fname1 , fname2 ,
            1 , 2 , "dct.tex" , 0 , 0 ) 
    && Detect_Base_Pointers_For ( fname1 , fname2 ,
            3 , 2 , "dctout.tex" , 0 , 0 ) 
    && Detect_Base_Pointers_For ( fname1 , fname2 ,
            4 , 3 , "ycc.tex" , 1 , 0 ) 
    && Detect_Base_Pointers_For ( fname1 , fname2 ,
            15 , 3 , "all.tex" , 0 , 0 ) )
    {
        printf ( "\nAll Ok.\n" ) ;
    } else {
        printf ( "\nStopped by error.\n" ) ;
    }
}

